/*
   $Id: Display.cpp 823 2017-04-12 06:35:33Z tfaraut $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/


#include <iostream>
#include <sstream>   // for ostringstream
#include <algorithm>  // std::reverse
#include <assert.h>
using namespace std;

#include "utils.h"
#include "Display.h"
#include "QueryBuffer.h"
#include "rtstack.h"

/**
 * \brief Display a vector of Hsp, grouped by "hits"
 *        - Sort the hsps: increasing query name, then decreasing scores
 *        - Create a HitPile
 *        - Store the hsps inside the HitPile (see AddSortedHspPtr for the details)
 *        - Sort the HitPile (see operator> for the Hit)
 *        - Call _DisplayHitPile
 * 
 * \param v A vector of Hsp to be printed
 * \param len The number of hsp to print: all if -1, none if 0, else the specified number (NOT USED !!!)
 */

void Display::DisplayHits(vector<Hsp*>& v,int len) const
{
	if ( v.empty() ) return;
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));

    HitPile hp(v);
    if ( _parameters.IsBwaLikeOutput() )
    {
		assert(hp.Size() == 1);
	}
    _DisplayHitPile(hp);
}

/**
 * \brief Display a HitPile object: call _DisplayHeader, DisplayQueryInfo, then display the hits
 * 
 * \param hp The HitPile to display
 * 
 */

void Display::_DisplayHitPile(const HitPile& hp) const
{
    Hsp* first_hsp = *(hp.begin()->hsps.begin());

    _DisplayHeader(first_hsp->GetQueryName(),_parameters.GetBankName());
    DisplayQueryInfo(first_hsp->GetQueryName());
    for (HitPile::const_iterator i=hp.begin(); i!=hp.end(); ++i) {
    	__DisplaySingleHit(*i);
    }
}

/**
 * \brief Display a hit: Call DisplaySubjectInfo, then display the hsps
 * \param h The hit to display
 * 
 */

void Display::__DisplaySingleHit(const Hit& h) const
{
	DisplaySubjectInfo(h.Name,h.score);
    for (vector<Hsp*>::const_iterator j=h.hsps.begin(); j != h.hsps.end(); ++j) {
            DisplayHsp(**j);
    }
}

/**
 * \brief The DisplayM2 class implements the methods specific to the M2 output format  (see documentation)
*/

void DisplayM2::DisplayTopHeader() const
{
	fprintf(_dsp_file.Get(),"%s\n",M2_HEADER);
}

void DisplayM2::DisplayHsp(const Hsp& hsp) const
{
	char sign=(hsp.GetStrand() > 0) ? '+' : '-';
	fprintf(_dsp_file.Get(),"%s\t%u\t%u\t%s\t%u\t%u\t%c\t%d\t%d\n",hsp.GetQueryName().c_str(),hsp.GetQueryStart(),hsp.GetQueryEnd(),hsp.GetSubjectName().c_str(),hsp.GetSubjectStart(),hsp.GetSubjectEnd(),sign,hsp.GetScore(),hsp.GetSize());	
}

/**
 * \brief The DisplayM3 class implements the methods specific to the M3 output format  (see documentation)
*/

void DisplayM3::DisplayTopHeader() const
{
	fprintf(_dsp_file.Get(),"%s\n",M3_HEADER);
}

void DisplayM3::DisplayHsp(const Hsp& hsp) const
{

    const char *char_sign=(hsp.GetStrand() >0) ? OUTPUT_PLUS:OUTPUT_MINUS;

	fprintf(_dsp_file.Get(),"%s\t%s\t",hsp.GetQueryName().c_str(),hsp.GetSubjectName().c_str());
	fprintf(_dsp_file.Get(),"[%u-%u]->[%u-%u](%u,%s,%d,,%2.0f%c)",hsp.GetQueryStart(),hsp.GetQueryEnd(),hsp.GetSubjectStart(),hsp.GetSubjectEnd(),hsp.GetSize(),char_sign,hsp.GetScore(),hsp.GetPIdent(),'%');	
	fprintf(_dsp_file.Get(),"[");
    hsp.GetMismatch().Print(_dsp_file.Get());
	fprintf(_dsp_file.Get(),"]");
	fprintf(_dsp_file.Get(),"{");
    hsp.GetQueryGaps().Print(_dsp_file.Get());
	fprintf(_dsp_file.Get(),"}");
	fprintf(_dsp_file.Get(),"{");
    hsp.GetSubjectGaps().Print(_dsp_file.Get());
	fprintf(_dsp_file.Get(),"}:");
	fprintf(_dsp_file.Get(),"\n");
}

/**
 * \brief The DisplayM8 class implements the methods specific to the M8 output format  (see documentation)
*/

void DisplayM8::DisplayHsp(const Hsp& hsp) const
{
	uint32_t s_start,s_end;
	if ( hsp.GetStrand() > 0 )
	{
		s_start = hsp.GetSubjectStart();
		s_end   = hsp.GetSubjectEnd();
	}
	else
	{
		s_start = hsp.GetSubjectEnd();
		s_end   = hsp.GetSubjectStart();
	}
	fprintf(_dsp_file.Get(),"%s\t%s\t",hsp.GetQueryName().c_str(),hsp.GetSubjectName().c_str());
	fprintf(_dsp_file.Get(),"%2.2f\t%d\t%d\t%u\t%u\t%u\t%u\t%d\t%s\t%d\n",hsp.GetPIdent(),hsp.GetSize(),hsp.GetMismatch().GetSize(),hsp.GetGapOpenings().GetSize(),hsp.GetQueryStart(),hsp.GetQueryEnd(),s_start,s_end,"1e-0",hsp.GetScore());
}

/**
 * \brief The DisplayM9 class implements the methods specific to the M9 output format  (see documentation)
*/

void DisplayM9::_DisplayHeader(const string& query_name, const string& bank) const
{
	fprintf(_dsp_file.Get(),"# glint rel $Rev: 823 $ - Compiled on %s at %s\n", __DATE__, __TIME__);
	fprintf(_dsp_file.Get(),"# Query: %s\n",query_name.c_str());
	fprintf(_dsp_file.Get(),"# Bank: %s\n",bank.c_str());
	fprintf(_dsp_file.Get(),"# Fields: Query id, Subject id, %c identity, alignment length, mismatches, gap openings, q. start, q. end, s. start, s. end, e-value, bit score\n",'%');
}

/**
 * \brief The DisplayM4 class implements the methods specific to the M4 output format  (see documentation)
*/
void DisplayM4::DisplayHsp(const Hsp& hsp) const
{
	cout<<hsp<<endl;
}

/**
 * \brief The DisplayM5 class implements the methods specific to the M5 output format  (see below)
*/
void DisplayM5::DisplayTopHeader() const
{
	fprintf(_dsp_file.Get(),"%s\n",M5_HEADER);
}

/**
 * \brief DisplayM5 : a dedicated output format with the additionnal information of query and subject sizes
 *   query qgaps qstart qend qsize subject sgaps sstart send ssize mismatch pident score strand
*/

void DisplayM5::DisplayHsp(const Hsp& hsp) const
{	
	char sign=(hsp.GetStrand() > 0) ? '+' : '-';
	fprintf(_dsp_file.Get(),"%s\t%d\t%u\t%u\t%u\t",hsp.GetQueryName().c_str(),hsp.GetQueryGaps().GetSize(),hsp.GetQueryStart(),hsp.GetQueryEnd(),hsp.GetQuerySize());
	fprintf(_dsp_file.Get(),"%s\t%d\t%u\t%u\t%u\t",hsp.GetSubjectName().c_str(),hsp.GetSubjectGaps().GetSize(),hsp.GetSubjectStart(),hsp.GetSubjectEnd(),hsp.GetSubjectSize());
	fprintf(_dsp_file.Get(),"%d\t%2.2f\t%i\t%c\n",hsp.GetMismatch().GetSize(),hsp.GetPIdent(),hsp.GetScore(),sign);
}

/**
 * \brief DisplayPsl : the Psl output format
 *   matches - Number of bases that match that aren't repeats
     misMatches - Number of bases that don't match
     repMatches - Number of bases that match but are part of repeats
     nCount - Number of "N" bases
     qNumInsert - Number of inserts in query
     qBaseInsert - Number of bases inserted in query
     tNumInsert - Number of inserts in target
     tBaseInsert - Number of bases inserted in target
     strand - "+" or "-" for query strand. For translated alignments, second "+"or "-" is for genomic strand
     qName - Query sequence name
     qSize - Query sequence size
     qStart - Alignment start position in query
     qEnd - Alignment end position in query
     tName - Target sequence name
     tSize - Target sequence size
     tStart - Alignment start position in target
     tEnd - Alignment end position in target
     blockCount - Number of blocks in the alignment (a block contains no gaps)
     blockSizes - Comma-separated list of sizes of each block
     qStarts - Comma-separated list of starting positions of each block in query
     tStarts - Comma-separated list of starting positions of each block in target
*/

void DisplayPsl::DisplayHsp(const Hsp& hsp) const
{
    char sign=(hsp.GetStrand() > 0) ? '+' : '-';
    
    uint16_t Ncounts = getNcounts(hsp);
    
    uint16_t qNumInsert  = hsp.GetQueryGaps().GetOpenings();
    uint16_t qBaseInsert = hsp.GetQueryGaps().GetSize();
    uint16_t tNumInsert  = hsp.GetSubjectGaps().GetOpenings();
    uint16_t tBaseInsert = hsp.GetSubjectGaps().GetSize();
   
    fprintf(_dsp_file.Get(),"%d\t%d\t%d\t%d\t",hsp.GetNumMatchingBases(), hsp.GetNumMismatch(), 0, Ncounts);
    fprintf(_dsp_file.Get(),"%d\t%d\t%d\t%d\t",qNumInsert,qBaseInsert,tNumInsert,tBaseInsert);
    fprintf(_dsp_file.Get(),"%c\t",sign);
    fprintf(_dsp_file.Get(),"%s\t%u\t%u\t%u\t",hsp.GetQueryName().c_str(), hsp.GetQuerySize(), hsp.GetQueryStart(), hsp.GetQueryEnd());
    fprintf(_dsp_file.Get(),"%s\t%u\t%u\t%u\t",hsp.GetSubjectName().c_str(), hsp.GetSubjectSize(), hsp.GetSubjectStart(), hsp.GetSubjectEnd());
    
    uint32_t nQuery = hsp.GetBQueryStart() - 1;
    uint32_t nBase  = (hsp.GetStrand()>0) ? hsp.GetSubjectStart()-1 : hsp.GetSubjectEnd();
   
    GapPile_const_iterator qgap = hsp.GetQueryGaps().begin();
    GapPile_const_iterator sgap = hsp.GetSubjectGaps().begin();
    
    Blocks q_blocks;
    Blocks s_blocks;
    bool was_gap = true;
    bool was_block = true;
    uint32_t q_start = nQuery;
    uint32_t s_start = nBase;
    uint32_t q_end, s_end;
    for (uint32_t i=0; i<hsp.GetSize(); i++ ) {
        if (qgap!=hsp.GetQueryGaps().end() && *qgap==i) {
            if (was_block) {
                q_end = nQuery;
                s_end = nBase;  
                q_blocks.push_back(Block(q_start,q_end+1));
                s_blocks.push_back(Block(s_start,s_end));
            }
            nBase+=hsp.GetStrand();
            ++qgap;
            was_gap = true;
            was_block = false;
        } else if (sgap!=hsp.GetSubjectGaps().end() && *sgap==i) {
            if (was_block) {
                q_end = nQuery;
                s_end = nBase;
                q_blocks.push_back(Block(q_start,q_end));
                s_blocks.push_back(Block(s_start,s_end));
            }
            ++sgap;
            nQuery++;
            was_gap = true;
            was_block = false;
        } else {
            if (was_gap) {
                q_start = nQuery;
                s_start = nBase;
            }
            nBase+=hsp.GetStrand();
            nQuery++;
            was_gap = false;
            was_block = true;
        }
    }
    q_blocks.push_back(Block(q_start,nQuery));
    s_blocks.push_back(Block(s_start,nBase));
    
    // In glint the query serves as the reference for alignments
    // In plus/minus strand we have to reverse the order of blocks
    if (hsp.GetStrand() < 0) {
        reverse(q_blocks.begin(),q_blocks.end());
        reverse(s_blocks.begin(),s_blocks.end());
    }
    
    std::ostringstream blocksizes;
    for (BlockIterator i = q_blocks.begin(); i != q_blocks.end(); ++i) {
        if (i != q_blocks.begin()) blocksizes<<",";  
        blocksizes << i->end - i->start + 1;
    }
    std::ostringstream qstarts;
    for (BlockIterator i = q_blocks.begin(); i != q_blocks.end(); ++i) {
        if (i != q_blocks.begin()) qstarts<<",";  
        if (hsp.GetStrand() > 0) {
            qstarts<< i->start;
        } else {
            qstarts<< hsp.GetQuerySize() - i->end;
        }
    }
    std::ostringstream sstarts;
    for (BlockIterator i = s_blocks.begin(); i != s_blocks.end(); ++i) {
        if (i != s_blocks.begin()) sstarts<<","; 
        if (hsp.GetStrand() > 0) {
            sstarts<< i->start;
        } else {
            sstarts<< i->end;
        }
    }
    uint16_t num_blocks = q_blocks.size();
    fprintf(_dsp_file.Get(),"%u\t%s\t%s\t%s\n",num_blocks,blocksizes.str().c_str(),qstarts.str().c_str(),sstarts.str().c_str());
}

uint16_t DisplayPsl::getNcounts(const Hsp& hsp) const {
    const unsigned char *nQuery = _query_ptr -> GetBase() + hsp.GetBQueryStart();
    const unsigned char *nBase  = __bank_seq + hsp.GetBSubjectStart();
    
    uint16_t Ncounts = 0;
    for (uint32_t i=hsp.GetQueryStart(); i<=hsp.GetQueryEnd(); i++, nQuery++ ) {
        if (toupper(*nQuery)=='N') Ncounts++;
    }
    for (uint32_t i=hsp.GetSubjectStart(); i<=hsp.GetSubjectEnd(); i++, nBase++ ) {
        if (toupper(*nBase)=='N') Ncounts++;
    }
    return Ncounts;
}

/**
 * \brief The DisplayBed class implements the methods specific to the Bed output format  (see documentation)
*/

void DisplayBed::DisplayHsp(const Hsp& hsp) const
{
	string chrom1 = hsp.GetSubjectName();
	uint32_t  start1 = hsp.GetSubjectStart() - 1;  // Bed files are 0-based
	uint32_t  end1   = hsp.GetSubjectEnd();
	char   strand1= hsp.GetStrand()>0? '+' : '-';	

	// bed formatting
	if ( !_parameters.IsPairEnd())
	{
		uint32_t       score  = hsp.GetScore();
		string name        = hsp.GetQueryName();
		fprintf(_dsp_file.Get(),"%s\t%i\t%i\t%s\t%i\t%c\n",chrom1.c_str(),start1,end1,name.c_str(),score,strand1 );
	}	
	// bedpe formatting
	else if ( hsp.IsPaired() ) // if not we are not printing anything
	{
		string name     = GetReadName( hsp.GetQueryName() );  //The fragment extension (/1 and /2) is removed if any
		HspPtr mate_ptr = hsp.GetMate();
		string chrom2   = mate_ptr -> GetSubjectName();
		uint32_t start2    = mate_ptr -> GetSubjectStart() - 1; // start is 0-based in bed format
		uint32_t end2      = mate_ptr -> GetSubjectEnd();       // end is 1-based in bed format
		//uint32_t score     = ((end1-start1)*hsp.GetScore() + (end2-start2)*mate_ptr -> GetScore()) / ( end1-start1 + end2-start2);
		uint32_t score     = hsp.GetScore() + mate_ptr -> GetScore(); //The score is simply the sum of the two scores
		char strand2      = mate_ptr -> GetStrand()>0 ? '+' : '-';
		fprintf(_dsp_file.Get(),"%s\t%i\t%i\t%s\t%i\t%i\t%s\t%i\t%c\t%c\n",chrom1.c_str(),start1,end1,chrom2.c_str(),start2,end2,name.c_str(),score,strand1,strand2 );	
	}
}

/**
 * \brief DisplayFasta - _DisplayHitPile calls DisplayHsp for the FIRST hsp only
 *
 * \param[in] hp The HitPile to print
 * 
 */

void DisplayFasta::_DisplayHitPile(const HitPile& hp) const
{	
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
		
	Hsp* h = *(hp.begin()->hsps.begin()); //The first hsp
	DisplayHsp(*h);
}

/**
 * \brief DisplayHsp - DisplayHsp only prints the fasta representation of the query sequence
 *
 * \param[in] h The Hsp to print
 * 
 */

void DisplayFasta::DisplayHsp(const Hsp & h) const
{	
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
		
	string SeqId = _query_ptr -> GetSeqId();
	fprintf(_dsp_file.Get(),">%s\tS=%d\tL=%d\tMM=%d\tG=%d\n",SeqId.c_str(),h.GetScore(),h.GetSize(),h.GetMismatch().GetSize(),h.GetQueryGaps().GetSize()+h.GetSubjectGaps().GetSize());
	fprintf(_dsp_file.Get(),"%s",_query_ptr -> GetSeqAsFasta() -> c_str());
}

/**
 * \brief DisplayFastq - _DisplayHitPile only calss DisplayHsp for the FIRST hsp
 *
 * \param[in] The HitPile to print
 * 
 */

void DisplayFastq::_DisplayHitPile(const HitPile& hp) const
{	
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
		
	Hsp* h = *(hp.begin()->hsps.begin()); //The first hsp
	DisplayHsp(*h);
}

/**
 * \brief DisplayFastq - DisplayHsp only prints the fastq representation of the query sequence
 *        NOTE - If the quality line is not conserved, we write a dummy quality just to be clean with the format
 *
 * \param[in] The HitPile to print
 * 
 */

void DisplayFastq::DisplayHsp(const Hsp& h) const
{	
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
		
	string SeqId =_query_ptr -> GetSeqId();
	fprintf(_dsp_file.Get(),"@%s\n",SeqId.c_str());
	fprintf(_dsp_file.Get(),"%s",_query_ptr -> GetSeqOneLine()->c_str());
	fprintf(_dsp_file.Get(),"%s","\n+\n");
	if ( _query_ptr -> GetQual() == NULL )
	{
		for (unsigned int i=0; i< _query_ptr -> GetSeqLength(); ++i)
		{
			fputc('!',_dsp_file.Get());
		}
	}
	else
	{
		const char* q = _query_ptr -> GetQual();
		for (unsigned int i=0; i< _query_ptr -> GetSeqLength(); ++i)
		{
			fputc(*q++ + '!',_dsp_file.Get());
		}
	}		
	fputc('\n',_dsp_file.Get());
}

/**
 * \brief Display the suitable regions of the query in fasta format
 *
 * \param[in] the frequentCodeBuffer 
 * \param[in] a reference on a vector with the suitable regions
 */
void DisplayFreq::DisplaySuitableFasta(const Buffer& freqBuffer, const vector<regionFreq>& suitableRegions) const
{
	string SeqId =_query_ptr -> GetSeqId();
	char * Seqbase  = (char*) _query_ptr -> GetBase();		
	char*  Codebase =  (char*) freqBuffer.GetBase();
	
	int cptr = 1;	
	for ( vector<regionFreq>::const_iterator rit = suitableRegions.begin(); rit != suitableRegions.end(); ++rit, cptr++ )
	{   
		char* nuc  = Seqbase + rit->start ;
		char* code = Codebase + rit->start -1;
		fprintf( _dsp_file.Get(),">%d|%s\tstart : %d\tend : %d\tsize : %d\n",cptr,SeqId.c_str(),rit->start,rit->end, rit->end-rit->start+1);
		for( uint32_t i = rit->start ; i <= rit->end ; ++i, nuc++, code++ )
		{	
			char c = ( *code == 0 ) ? *nuc : 'N';
			fprintf(_dsp_file.Get(),"%c",c);
			if ( (i- rit->start+1)%60 == 0 )
				fprintf(_dsp_file.Get(),"\n");
		}
		if ( (rit->end-rit->start+1)%60 != 0)
			fprintf(_dsp_file.Get(),"\n");	
	}			
}
