/*
   $Id: hsp.h 822 2017-04-11 17:07:04Z tfaraut $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef __HSP_H__
#define __HSP_H__

#include <cmath>

#include "Parameters.h"
#include "SeqInfo.h"

/**
 * \brief A GapPile is a "stack" of gaps. It encapsulates a vector of int16_t (the gap positions, 0 = first position in the hsp)
 *        - A new gap can be added with the Push function
 *        - All already stored gap can be shifted with the Increment function
 *        - The gaps can be sorted with the Sort() function
 *        - With begin() and end(), you can iterate through the object (const iterators only)
**/

typedef uint32_t Gap;
typedef vector<Gap>::const_iterator GapPile_const_iterator;

class GapPile {
    public:
	    GapPile(): __sorted(false) {};
	    void Push(Gap g);
	    void Increment(Gap l);
	    void Clear();
		GapPile Reverse(uint32_t) const;
	    void Print(FILE *pf) const;
	    uint16_t GetSize() const;
        uint16_t GetOpenings() const;
	
	    typedef vector<uint32_t>::const_iterator const_iterator;
	    const_iterator begin() const;
	    const_iterator end() const;
	
	    void PrintDebug(FILE *pf);
	
		friend class AlignEngine;
        friend class Hsp;
        
    private:
	    mutable vector<Gap> gaps; // mutable for the Sort() method
	    mutable bool __sorted;
	    void __Sort() const;      // When sorted, or reversed, the GapPile is NOT considered as modified
};


/**
 *  \brief Block structure : a Block is an ungapped alignment
 * 
*/
typedef struct Block {
    uint32_t start, end;
    Block(uint32_t s,uint32_t e): start(s),end(e) {};
    void Print() const { fprintf(stderr,"%d\t%d\n",start,end); };
} Block;

typedef vector< Block > Blocks;
typedef vector< Block >::const_iterator BlockIterator;

/**
 * \brief The well-known Highest Segment Pair concept : a local alignment between a query and a subject 
 *        with the following information:
 *        - q_start, q_end: the start, end positions of the query   ( 0 = base of the buffer in the memory )
 *        - s_start, s_end: the start, end positions of the subject ( 0 = base of the buffer in the memory )
 *        - The strand (-1, 1)
 *        - a score, the size, etc.
 *        - Several GapPiles: gaps on query, on subject, mismatches, and gap openings
*/

typedef pair< string, uint32_t >  Tag; 
typedef vector< Tag > MapInfo;
typedef MapInfo* MapInfoPtr;
typedef vector< Tag >::const_iterator MapInfoIterator;
class Hsp;
typedef Hsp* HspPtr;



class Hsp {
	
	public:
	    Hsp():     
					q_start(0),
					q_end(0),
					s_start(0),
					s_end(0),	
					__strand(1),
	               __score(0),
	               __size(0), 
	               __excluded(false), 
	               __paired(false), 
	               __mapquality(0), 
	               __mate(NULL),  
	               __mappinginfo() 
	               {};
	    
	    friend class HspPile;
	    friend bool operator<(const Hsp& p, const Hsp& q);
	    friend bool CompHspSubjPosPtr(const Hsp* p, const Hsp* q);
	    friend bool CompHspScorePtr(const Hsp* p, const Hsp* q); 
	    friend bool CompHspQuePtr(const Hsp* p, const Hsp* q); 
		friend ostream & operator<<(ostream& os, const Hsp &);
	    friend class HitPile;
		friend class AlignEngine;
		friend class AlignEngineSelf;
		friend class QueryBufferSam;
		friend class GlintEngineDouble;
		friend class GlintEngineDoublePCR;
		
		const GapPile& GetQueryGaps()   const { return q_gap; };
		const GapPile& GetSubjectGaps() const { return s_gap; };
		const GapPile& GetGapOpenings() const { return gap_opening; };
		const GapPile& GetMismatch()    const { return mismatch; };
	
		const string& GetQueryName()    const { return Query.Name; };
		const string& GetSubjectName()  const { return Subject.Name; };
		uint32_t GetQuerySize()   const { return Query.Size(); };             // The size of the Query
		uint32_t GetSubjectSize() const { return Subject.Size(); }            // The size of the Subject
		uint32_t GetQueryLength()   const { return q_end - q_start + 1; };   // The size of the hsp on the Query
		uint32_t GetSubjectLength() const { return s_end - s_start + 1; };   // The size of the hsp on the Subject
		
		uint32_t GetSize()          const { return __size; };
		uint32_t GetScore()         const { return __score; };
		int8_t  GetStrand()        const { return __strand; };
		
		// Retrieve the start,end on the query, on the subject, relative to the sequence (1 = 1st pair)
		uint32_t GetQueryStart()    const { return q_start - Query.start + 1; };
		uint32_t GetQueryEnd()      const { return q_end - Query.start + 1; };
		uint32_t GetSubjectStart()  const { return s_start - Subject.start + 1; };
		uint32_t GetSubjectEnd()    const { return s_end - Subject.start + 1; };
		
		// Retrieve the start,end on the query, on the subject, relative to the buffer (0 = 1st char in the buffer, ie a separator)
		uint32_t GetBQueryStart()   const { return q_start; };
		uint32_t GetBQueryEnd()     const { return q_end; };
		uint32_t GetBSubjectStart() const { return s_start; };
		uint32_t GetBSubjectEnd()   const { return s_end; };
                
		bool IsPaired() const { return __paired; };
		HspPtr GetMate() const { return  __mate; };
		uint8_t GetFragNum() const { return __fragnum; };
		uint16_t GetMapQuality() const { return  __mapquality; };
		
		void SetMapQuality(uint16_t mapqual) { __mapquality = mapqual; };
		
		const MapInfo& GetMapInfo() const { return  __mappinginfo ; };
		
		void AddMapInfo( Tag t ) { __mappinginfo.push_back( t ); };
		uint32_t GetMapInfoValue( string label ) 
		{ 	
			MapInfoIterator i =  __mappinginfo.begin();
			while ( i!=  __mappinginfo.end() && i->first != label ) i++;
			if ( i != __mappinginfo.end() )
			{
				return i->second;
			}
			else
			{
				throw( logic_error("ERROR - "+label+" is not an mapping info label") );
			}
		
		};
		
		// VIRER CETTE METHODE - CF. Display.cpp L144
		SequencePosition& GetQueryInfo() { return Query; };
		
	    // Return the edit distance to the ref seq (cf. SAM documentation)
	    int GetEditDist() const { return mismatch.GetSize()+q_gap.GetSize()+s_gap.GetSize(); };
        
        // Return the number of matching bases
        int GetNumMatchingBases() const { return GetSize()-GetEditDist(); };
        
        // Return the number of mismatches
        int GetNumMismatch() const { return mismatch.GetSize(); };
	    
	    // Return the %age identity
	    float GetPIdent() const { return 100*(1-(float)(GetEditDist())/GetSize()); };
	    
	    //Pairing two hsps
	    friend void Pair(HspPtr first_ptr, HspPtr second_ptr ) 
	    { 
			first_ptr->__paired  = true; 
			second_ptr->__paired = true;
			first_ptr->__mate    = second_ptr;
			second_ptr->__mate   = first_ptr;
		};	        
		// Clear the Hsp 
		void Clear() 
		{
			q_gap.Clear();
			s_gap.Clear();
			gap_opening.Clear();
			mismatch.Clear();
			__mappinginfo.clear();
			__paired     = false;
			__mate       = NULL;
			__mapquality = 0;
		}
		
		//Update query information for empty hsp
		void UpdateQueryInfo(const char* name,int64_t length)
		{
			Subject.Name = "";
			Query.Name  = name;
			Query.start = 0;
			Query.end	= length;
			q_start  = 0;
			q_end    = length;	
		}
		
#ifdef DEBUGLOG
		void LogHsp( const char*, const char * );
#endif

	private:	
		uint32_t q_start, q_end, s_start, s_end;
		int32_t min_diag,max_diag;
		int8_t  __strand;
		uint32_t __score;
		uint32_t __size;
	    GapPile q_gap, s_gap, gap_opening, mismatch;
	    
	    SequencePosition Query;
	    SequencePosition Subject;
	          
		bool __excluded;
		bool __paired;
		
		//For computing the mapping quality 
		uint16_t __mapquality;
		
		//For pair-end sequences
		HspPtr __mate;
		uint8_t  __fragnum;
	
		MapInfo __mappinginfo;
	
		// For debugging purpose only
		uint16_t seed_number;
		uint32_t maxscore;
};


/**
 * \brief An HspPile encapsulates a lot of \b pre-allocated (for perf) Hsp objects.
 *        The number of Hsp is passed by the constructor.
 *        - GetTop may be used to retrieve the highest stored Hsp
 *        - Commit is used to "record" the top Hsp in the object, and to propose a new top
 *        - Lookup is used to check overlaps between  new hsps and already committed hsps
 *        - When committing, if every preallocated is committed, AND if we are working in hsp_in_the_fly mode, the hsp are printed
 */

class match;
class Display;
class ComparisonBase;
typedef vector<Hsp*>::iterator HspIterator;            // const_iterator SERAIT PLUS LOGIQUE, MAIS ON PEUT PAS
typedef pair<Hsp*,Hsp*> MatePair;
typedef vector<MatePair>::iterator MatePairIterator;

class QueryBuffer;
class HspPile {
    public:

		HspPile(Display& d);
		~HspPile();
		
		/* Commit increments top, WITHOUT increasing the size of p
		If top is at the highest, p and h_list are dumped and cleared */
		
		void Commit(const QueryBuffer* q=NULL);
		void SwitchPiles();
		
		void Sort();		
		void SortQuery();
		
		/*void SortQue();
		void SortSub();
		void SortSub1();*/
		//void PrintDebug(FILE *pf);  //For debugging purpose
		
		void PrintDebugNumbering(const vector<Hsp*>& v,int len,FILE *pf) const;
		void Reset();
		void PrintDebugDisplay();
		bool Lookup (const match&);
		bool Lookup (const Hsp&);
		
		void PrintOuput();
#ifdef DEBUGLOG
		void LogPile( const char* );
#endif

	 /* For debugging purpose only  */

	 void PrintSortedPile();
	 void PrintPile();

    /* return the top of the pile, slightly clearing the returned hsp */
    Hsp& GetTop();
    
    /* return the number of hsp stored in the pile */
    int GetNbHsp() const { return top; };

    HspIterator begin();
    HspIterator end();

	HspIterator beginhsp();
    HspIterator endhsp();

    private:
		const Parameters& __parameters;
		bool _LookupCurrent(int32_t, uint32_t);
		bool _LookupPrevious(uint32_t, uint32_t);
		void _Insert(Hsp*);         // Insert an hsp in the list hsp_list
		uint16_t top;
		bool hsp_prev_empty;        // if true, there is nothing interesting in hsp_prev
		vector<Hsp*>*hsp_prev;      // The 'old' preallocated hsps
		vector<Hsp*>*hsp_current;   // The 'current' preallocated hsps
		list<Hsp*> hsp_list;        // The committed hsps, sorted as specified by _Insert
		list<Hsp*> hsp_valid;       // The valid hsps, selected by DetectOverlappingHsps
		//list<chain*> hsp_chains;  // The chained hsps, as constructed by Chaining
		
		Display& display;
};

/* The inline methods of those classes */

inline void GapPile::Push(uint32_t g) {gaps.push_back(g);__sorted=false;}
inline void GapPile::Increment(uint32_t l) { for(uint16_t i=0;i<gaps.size();i++) gaps[i]+=l;}
inline void GapPile::Clear() {gaps.clear();}
inline void GapPile::__Sort() const { if (!__sorted){sort(gaps.begin(),gaps.end());__sorted=true;};}
inline void GapPile::Print(FILE *pf) const
{
    for(uint16_t i=0;i<gaps.size();i++)
    {
        if (i>0) fprintf(pf,",");
        fprintf(pf,"%d",gaps[i]);
    }
    fprintf(pf,"\n");
}
inline uint16_t GapPile::GetSize() const { return gaps.size();}
inline GapPile::const_iterator GapPile::begin()   const {__Sort();return gaps.begin();}
inline GapPile::const_iterator GapPile::end()     const {__Sort();return gaps.end();}


inline uint16_t GapPile::GetOpenings() const 
{
    uint16_t num = 0;
    int32_t prev = 0;
    for(uint16_t i=0;i<gaps.size();i++) {
        if (abs(int32_t(gaps[i])-prev)>1) num++;
        prev = int32_t(gaps[i]);
    }
    return num;
}

inline void GapPile::PrintDebug(FILE *pf)
{
    fprintf(pf,"gap pile \n");
    for(uint16_t i=0;i<gaps.size();i++) {
        fprintf(pf,"%d \t",gaps[i]);
    }
    fprintf(pf,"\n");
}

/*
      The hsp comparison operator and functions
*/

bool operator<(const Hsp& p, const Hsp& q);
inline bool CompHspPtr(const Hsp* p, const Hsp* q) {return *p<*q;}
inline bool CompHspScorePtr(const Hsp* p, const Hsp* q) {return p->GetScore() > q->GetScore();}
inline bool CompHspQuePtr(const Hsp* p, const Hsp* q) { return 	(p->q_start<q->q_start); }
inline bool CompPairHsp(const MatePair& p, const MatePair& q) { return (p.first->GetScore()+p.second->GetScore()) > (q.first->GetScore()+q.second->GetScore()); }

/*
     An alternate comparison function for Hsp
		1st key: Name (i.e chromosome increasing order)
		2st key: subject position
*/
inline bool CompHspSubjPosPtr(const Hsp* p, const Hsp* q) 
{
	if (p->Subject.Name==q->Subject.Name)
        return (p->s_start < q->s_start);
    else
        return (p->s_start < q->s_start);	
}

inline void HspPile::Sort() {sort(hsp_current->begin(),hsp_current->begin()+top,CompHspPtr);}
inline void HspPile::SortQuery() {sort(hsp_current->begin(),hsp_current->begin()+top,CompHspQuePtr);}

/*
inline void HspPile::SortSub1() {sort(hsp_current->begin(),hsp_current->begin()+top,CompHspSubPtr);}
inline void HspPile::SortQue() {hsp_valid.sort(CompHspQuePtr);}
inline void HspPile::SortSub() {hsp_valid.sort(CompHspSubPtr);}
*/ 
inline void HspPile::Reset()  { top=0; hsp_list.clear();}
inline Hsp& HspPile::GetTop() {
    Hsp *h=(*hsp_current)[top];
    h->Clear();
    return *h;
}

inline HspIterator HspPile::begin() {return hsp_current->begin();}
inline HspIterator HspPile::end()   {return hsp_current->begin()+top;}

inline void HspPile::_Insert(Hsp* h)
{
    for (list<Hsp*>::iterator i=hsp_list.begin(); i != hsp_list.end(); ++i)
    {
        Hsp ii = **i;
        /* Si la diagonale est superieure a la derniere diagonale ou si
           sur la meme diagonale que le premier element de la liste mais avec une position plus en aval on insere  */
        if (h->max_diag > ii.max_diag || (h->max_diag == ii.max_diag && h->s_end > ii.s_end)) {
            hsp_list.insert(i,h);
            return;
        };
    };
    hsp_list.insert(hsp_list.end(),h);
}

/**
 * \brief A Hit is a set of hsps belonging to the same query sequence
 *          - Name is the name of the query sequence
 *          - score is the best hsp score
 *          - A vector of hsps is wrapped inside this structure
*/

struct Hit {
	    Hit(const string& n="", int16_t s=0) : score(s),Name(n) {};

		uint32_t score;
		string Name;
	    vector<Hsp*> hsps;
};

/**
 * \brief Use for sorting the hits: the key is the score, the hits are sorted in decreasing order of score
 */

inline bool operator< (const Hit& p, const Hit& q) { return p.score > q.score; }

/**
 * \brief A HitPile manages a list of hits
 *         This object is built from a sorted vector of Hsp*
 * \todo ca n'est pas du tout une Pile !
 */

class HitPile {
    public:
	    HitPile (vector<Hsp*>& v);
		uint32_t Size() const { return __hits.size(); };
	
	    /* Iterating through the hits */
	    typedef list<Hit>::const_iterator const_iterator;
	    const_iterator begin() const   {return __hits.begin();};
	    const_iterator end() const     {return __hits.end();};

    private:
		const Parameters& __parameters;
		list<Hit> __hits;
	    void __AddSortedHspPtr(Hsp* h);
		void __NewHit ( Hsp* hsp_ptr)
		{
			Hit hit ( hsp_ptr->Subject.Name, hsp_ptr->GetScore());
			__hits.push_back( hit );
		}
};

#endif
