/*
   $Id: masque.cpp 713 2013-06-17 14:54:05Z tfaraut $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

/*===========================================================================
                          masque  -  description
                             -------------------
    begin				: septembre 2005
	 objet				: Everything for the definition and initilization of the Mask : see description ine the main

=============================================================================*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
using namespace std;
#include <iostream>
#include "ptypes.h"
#include "utils.h"
#include "masque.h"
//#include "memory.h"

#include "word.h"

#include "rtstack.h"



/** 
 * \brief Initializing the Mask - Because of the singleton semantics, the Mask objects are initialized AFTER their construction
 *        However, the initialized boolean member tracks the initializations and refuses a second initialization
 */

void Mask::Initialize (const string& str_mask) throw (rtstack)
{
	if (initialized)
	{
		// If already initialized with the same parameters, return
		if ( str_mask == string_mask )
		{
			return;
		}
		// If already initialized with other parameters, it is an error
		else
		{
			string msg = "ERROR - the Mask singleton is already initialized\n";
			msg += ERROR_FILE_LINE(__FILE__,__LINE__);
			throw(rtstack(msg));
		}
	}
	else
	{
		initialized = true;
	}
	
	int j,k;
	size_t i;
	uint32_t b;
	/*int a[2];   Warning, set but not used
	a[0]=1;
	a[1]=0;*/

	//cerr<<"Here init stringmask "<<str_mask<<"\n";

	if (str_mask=="")
    {
        length = 0;
        string_mask = "";
    }
    else
    {
        string_mask=str_mask;
        length=string_mask.length();
    };

    if (length == 0)
    {
        pos = NULL;
        leaf_level = 0;
        level_pos = NULL;
        return;
    };

    /* Initialize the level_length array */
	for (i=0;i<MAX_LEAF_LEVEL;i++) {
		level_length[i]=0;
	}

    /* Create and init the pos and level_length arrays, also leaf_level */
    pos=(short*) MAlloc(length*sizeof(short));
	int min=length;
	int max=0;
	for (i=0;i<length;i++) {
		pos[i]=str_mask[i]-'1' + 1;            /* converts char 1 (2,3,...) to the corresponding integer value*/
		min=(pos[i]<min && pos[i]!=0) ? pos[i] : min;   /* Here allowing for 0 in the mask   */
		max=(pos[i]>max) ? pos[i] : max;
		level_length[pos[i]]++;
	};
    leaf_level=max;
    
    min_level = min;
    max_level = max;

    /* Check for consistency */
	if (min_level != ROOT_LEVEL || max_level > MAX_LEAF_LEVEL)
	{
		ostringstream msg;
		msg << "ERROR - The Levels ([" << min_level << "-" << max_level << "]) of the mask ";
		msg << str_mask << " are out of range ([1-" << MAX_LEAF_LEVEL << "])\n";
		msg << ERROR_FILE_LINE(__FILE__,__LINE__);
		throw (rtstack(msg.str()));
	}

    /* We refuse a mask with level_length higher than 4, except for the root: if level_length is more than 4, the corresponding
       code has more more thant 8 bytes, and there will be some overflow in the index binary file */

    for (i=ROOT_LEVEL+1; i<=leaf_level; i++)
    {
        if (level_length[i] > 4) 
        {
			ostringstream msg;
			msg << "ERROR - The Level " << i << " has a length higher than 4 (" << level_length[i] << ")\n";
			msg << ERROR_FILE_LINE(__FILE__,__LINE__);
			throw (rtstack(msg.str()));
		}
	}

    //bm=(uint64_t*)memory_survey((max+1)*sizeof(uint64_t),(char*)"unsigned long*");
    bm = (uint64_t*) MAlloc( (max+1)*sizeof(uint64_t) );

	for (j=min;j<=max;j++)
	{
		bm[j]=0;
		for (i=0;i<length;i++)
		{
			b=(pos[i]<=j) ? 1 : 0;
			bm[j]<<=1;
			bm[j]|=b;
		}
	}
	//fprintf(stderr,"le masque est de taille %d : min=%d  max=%d  %d  \n",length,min,max,leaf_level);
	level_pos = (short**) MAlloc ( (max_level+1)*sizeof(short*) );

	/*fprintf(stderr,"Le masque est %s \n",pos);
	*/

	for (j=min_level;j<=max_level;j++)
	{
		level_pos[j] = (short*) MAlloc ( level_length[j]*sizeof(short) );
		k=0;
		for (i=0;i<length;i++) {
			if (pos[i]==j) level_pos[pos[i]][k++]=i;
		}
	}
}


Mask::~Mask()
{
	for (int j=min_level;j<=max_level;j++)
	{
		free ( level_pos[j] );
	}
	free ( level_pos );
	free ( bm );
	free ( pos );
}

/*
  Name:      Mask::Reconstruct
  Usage:     string s = m.Reconstruct(s,h_code,level)
  Function:  Partially reconstruct the word
  Arg:  s      ==> The word to be reconstructed (input/output)
        h_code ==> The hash code used for reconstruction
        level  ==> The mask level of the reconstruction
  Return: s
*/

string Mask::Reconstruct (const string & w,uint32_t l,int level) const
{
    string mot=w;
	if (level==ROOT_LEVEL) {
        mot = "";
		for (int k=0;k<length;k++)
            mot += '.';
	}
	else
	{
		mot = ToLower(mot);
	}
	for (int k=0;k<level_length[level];k++) {
		unsigned long c=l & 3;
		l>>=2;
		int i = level_pos[level][level_length[level]-k-1];
		mot[i] = Word::decode(c);
	}
	//if (level==m->leaf_level) mot->m[m->taille]='\0';
	return mot;
}

ostream & operator<<(ostream& os, const Mask& m)
{
	os << m.string_mask;
/*
    for (int i=0;i<m.length;i++)
        os << m.pos[i];
    os << "\n\n";
    */
    return os;
}

